structure CodeGen = 
struct 
open DataTypes;
open SymbolTable;
open PushCount;
open TypeCheck;

exception NotImplemented;
exception UnreadableType;

fun conjoin(codes) = List.foldr (op ^) "" codes;

(* Yeah.  This code is stateful.  So sue me. *)
val label_num = ref 0;

val class_name = ref "";

fun gen_label() =  (
      label_num := !label_num + 1;
      "label" ^ Int.toString(!label_num)
    )

val jvm_register = ref 0

fun allocate_jvm_var() = 
     let val rv = !jvm_register in (
       jvm_register := !jvm_register + 1;
     rv 
     )
     end

fun typeof(Checked(t, expr)) = t
  | typeof(_) = raise Halt

fun init_type(Const(t)) = t
  | init_type(t) = t

fun emit("") = ""
  | emit(code) = code^"\n";

fun emit1(opcode, operand) = 
         emit(opcode^" "^operand);

fun emit2(opcode, operand1, operand2) = 
         emit(opcode^" "^operand1^" "^operand2);

fun emit3(opcode, operand1, operand2, operand3) = 
         emit(opcode^" "^operand1^" "^operand2^" "^operand3);

fun emit_label(label) = emit1(label, ":")

fun goto(label) = emit1("goto", label)

fun emit_all(generator, generator_args) = 
    List.foldr (op ^) "" (List.map(generator) generator_args)

fun emitLabel(label) = emit(label^":")

fun ldc(x) = emit1("ldc", x)

fun istore(x : int) = emit1("istore", Int.toString(x))

fun iload(x : int) = emit1("iload", Int.toString(x))

fun ld_int(x) = 
      if x < 0 then
        ldc("-"^hd(tl(String.fields (fn x => x = #"~") (Int32.toString(x)))))
      else
        ldc(Int32.toString(x))
fun ifeq(label) = emit1("ifeq", label)

fun iconst_0() = emit("iconst_0")

fun iconst_1() = emit("iconst_1")

fun is_static(name, table) = 
    case get_symb(name, table) of CodegenInfo({info : symb_info, addr = StaticField}) => true
                               |  _                                                   => false

fun get_addr(name, table) = 
    case get_symb(name, table) of 
           CodegenInfo({info : symb_info, addr : jvm_address }) => addr 
       | _ => raise Halt

fun get_var(name, table) = 
    case get_addr(name, table) of StackVar(x) => x 
                                | _ => raise Halt 

fun name_of(Name(id(name), _, _)) = name 
  | name_of(ArrayDeref(id(name), _, _, _)) = name
  | name_of(Checked(expr, e)) = name_of(e) 
  | name_of(_) = raise Halt


fun put_static(tipe, name) = emit2("putstatic", !class_name ^ "/" ^ name, jvm_type(tipe));
fun get_static(tipe, name) = emit2("getstatic", !class_name ^ "/" ^ name, jvm_type(tipe));
fun call(name, "", return_type) = emit1("invokestatic", !class_name ^ "/" ^ name^"()"^jvm_type(return_type))
  | call(name, jasmin_arg_decls, return_type) = emit1("invokestatic", !class_name ^ "/" ^ name^"("^jasmin_arg_decls^")"^jvm_type(return_type))

fun gen(Program(id(name), members)) =  (
      class_name := name;
      let val table = open_scope(make_symbol_table())  in (
       conjoin([
        emit1(".source", name^".csx"),
        emit2(".class", "public", name),
        emit1(".super", "java/lang/Object"),
        gen_members(members, insert_symb(name, CSXClass(name), table)) ]))
      end
      )
and 
    gen_members(MemberDecls(fields, methods), table) = 
       conjoin([emit_all((fn f => gen_field_decl(f, table)), fields),
                emit_all((fn m => gen_decl(m, table)), methods),
                cl_init(fields, table)
               ])
and 
    cl_init(fields, table) = 
       conjoin([
         emit2(".method", "static", "<clinit>()V"),
         emit2(".limit", "stack", Int.toString(push_count_decls(fields))),
         emit2(".limit", "locals", Int.toString(0)),
         init_fields(fields, table),
         emit("return"),
         emit1(".end", "method")
       ])
and 
    init_fields(fields, table) = 
        conjoin(List.map (fn f => init_field(f, table)) fields)
and
    init_field(Field(Array(tipe, n), id(s), _, _), table) = 
       conjoin([
         ldc(Int32.toString(n)),
         emit1("newarray", jasmin_type(tipe)),
         put_static(Array(tipe, n), s)
       ])
  | init_field(Field(tipe, id(s), _, _), table) = ("") 
  | init_field(InitializedField(tipe, id(s), value, _, _), table) = conjoin([gen_expr_t(value, table), put_static(init_type(tipe), s)])
  | init_field(Method(tipe, id(s), args, fields, stmts, _, _), table) = (raise Halt)
  | init_field(MainMethod(decls, stmts, _, _), table) = (raise Halt)
  | init_field(FormalArg(tipe, id(s), _, _), table) = (raise Halt)
  | init_field(NoneDecl, table) = ("")
and
     gen_exprs(exprs, table) = conjoin(List.map (fn e => gen_expr_t(e, table)) exprs)
and 

      gen_typed_expr(NoneExpr, tipe, table) = ""

    | gen_typed_expr(ArithmeticBinaryOperation(lhs, arit_binary_operator, rhs, line, col), tipe, table) = 
         conjoin([gen_expr_t(lhs, table), gen_expr_t(rhs, table), gen_arit_binary_op(arit_binary_operator, table)])

    | gen_typed_expr(LogicalBinaryOperation(lhs, logi_binary_operator, rhs, line, col), tipe, table) = 
         conjoin([gen_expr_t(lhs, table), gen_expr_t(rhs, table), emit(gen_logi_binary_op(logi_binary_operator, table))])

    | gen_typed_expr(RelationalBinaryOperation(lhs, rel_binary_operator, rhs, line, col), tipe, table) = 
         cond_binary(lhs, rhs, gen_rel_binary_op(rel_binary_operator, table), table)

    | gen_typed_expr(EqualityBinaryOperation(lhs, equal_binary_operator, rhs, line, col), tipe, table) = 
         cond_binary(lhs, rhs, gen_equal_binary_op(equal_binary_operator, table), table)

    | gen_typed_expr(UnaryOperation (NOT, expr, line, col), tipe, table) = 
         let val false_label  = gen_label()
             val end_label =  gen_label() in
           conjoin([gen_expr_t(expr, table),
                    ifeq(false_label),
                    iconst_0(),
                    goto(end_label),
                    emit_label(false_label),
                    iconst_1(),
                    emit_label(end_label)
                   ])
         end

    | gen_typed_expr(UnaryOperation (Cast(target_tipe), expr, line, col), tipe, table) = 
       let val alreadyBoolean = gen_label() in
       conjoin([
         gen_expr_t(expr, table),
         case target_tipe of Char=> conjoin([ldc(Int.toString(127)), emit("iand"), emit("i2b")])
                           | Bool => conjoin([
                               emit("dup"),
                               emit1("ifeq", alreadyBoolean),  
                               emit("pop"),
                               ldc(Int.toString(1)),
                               emit_label(alreadyBoolean)])
                           | Int => ""
                           | _ => ""
         ])
       end
    | gen_typed_expr(CharLit("'\\n'", line, col), tipe, table) = (
       ldc(Int.toString(Char.ord(#"\n"))))
    | gen_typed_expr(CharLit("\\t", line, col), tipe, table) = (
       ldc(Int.toString(Char.ord(#"\t"))))
    | gen_typed_expr(CharLit(the_char, line, col), tipe, table) =  (
      ldc(Int.toString(Char.ord(String.sub(the_char, 1))))
    )
    | gen_typed_expr(IntLit(the_num, line, col), tipe, table) = 
     (ld_int(the_num))
    | gen_typed_expr(StrLit(the_str, line, col), tipe, table) = ldc(the_str)
    | gen_typed_expr(True(_), tipe, table) = emit("iconst_1")
    | gen_typed_expr(False(_), tipe, table) = emit("iconst_0")
    | gen_typed_expr(ArglessFnCall(id(name), line, col), tipe, table) = call(name, "", tipe)
    | gen_typed_expr(FnCall(id(name), actual_args, line, col), tipe, table) = (
      let val jasmin_arg_decls = 
         if List.length(actual_args) > 2 then
              conjoin(List.map (fn x => jvm_type(typeof(x))^"") actual_args)
         else
              conjoin(List.map (fn x => jvm_type(typeof(x))) actual_args) in
         conjoin([
           gen_actual_args(actual_args, table),
           call(name, jasmin_arg_decls, tipe)
         ])
      end
      )
    | gen_typed_expr(Name(id(name), line, col), tipe, table) =  (
         if is_static(name, table) then
            get_static(tipe, name)
         else
           let val var = get_var(name, table) in (
             case tipe of Array(t, n) => array_load(var)
                        | ArrayParam(t) => array_load(var)
                        | _ => iload(var)
             )
           end
    )

    | gen_typed_expr(ArrayDeref(id(name), expr, line, col), tipe, table) = 
       (* if its a field, get static, else aload *)
       (* in any case, visit the subscript and iaload, baload or caload based on the subscript *)
       let val static = is_static(name, table)
           val dummy = 1
           val load_inxion = if static then 
                                get_static(Array(tipe, dummy), name) 
                             else  
                                emit1("aload", Int.toString(get_var(name, table)))
           val deref_inxion = case tipe of Int => "iaload"
                                         | Char => "caload"
                                         | Bool => "baload"
                                         | _ => raise Halt
       in
        conjoin([
         emit(load_inxion), 
         gen_expr_t(expr, table),
         emit(deref_inxion)
         ])
       end
    | gen_typed_expr(ActualArgument(expr, line, col), tipe, table) = gen_expr_t(expr, table)
    | gen_typed_expr(Checked(tipe0, expr), tipe1, table) = (
       gen_expr_t(expr, table)
    )
and 
     cond_binary(lhs, rhs, operator_code, table) = 
      let val true_label = gen_label()
          val end_label = gen_label() in
         conjoin([gen_expr_t(lhs, table), 
                  gen_expr_t(rhs, table), 
                  emit1(operator_code, true_label),
                  iconst_0(),
                  goto(end_label),
                  emitLabel(true_label),
                  iconst_1(),
                  emitLabel(end_label)
                  ])
      end
and
     gen_expr_t(Checked(tipe, expr), table) = 
       let val x = gen_typed_expr(expr, tipe, table) in
         x
       end
     | gen_expr_t(_, table) = (print("Compiler bug: We cannot gen codes for unchecked expressions\n"); raise Halt)
and 
      gen_actual_args(args, table) = conjoin(List.map (fn e =>(gen_expr_t(e, table))) args)
and 
      gen_decls(decls, table) = conjoin(List.map (fn d => gen_decl(d, table)) decls)
and 
      gen_fields(decls, table) = conjoin(List.map (fn d => gen_field_decl(d, table)) decls)
and
      gen_field_decl(Field(Array(tipe, size), id(s), line, col), table) = (
        insert_symb(s,
                CodegenInfo({info = Info({name = s, 
                                         tipe = tipe, 
                                         decl = Field(Array(tipe, size), id(s), line, col), 
                                         whatisit = CSXField}),
                             addr = StaticField}), 
                    table);

        emit2(".field", "public static", s ^ " " ^ jvm_type(Array(tipe, size)))
    )

    | gen_field_decl(Field(tipe, id(s), line, col), table) = (
        insert_symb(s,
                CodegenInfo({info = Info({name = s, 
                                         tipe = tipe, 
                                         decl = Field(tipe, id(s), line, col), 
                                         whatisit = CSXField}),
                             addr = StaticField}), 
                   table);
        emit2(".field", "public static", s ^ " " ^ jvm_type(tipe))
    )
    | gen_field_decl(InitializedField(UnknownConst, id(s), value, line, col),
    table) = (print("gen_field_decl(InitializedField(UnknownConst...))"); raise Bug)

    | gen_field_decl(InitializedField(tipe, id(s), value, line, col), table) =(
        insert_symb(s,
                CodegenInfo({info = Info({name = s, 
                                         tipe = tipe, 
                                         decl = Field(tipe, id(s), line, col), 
                                         whatisit = CSXField}),
                             addr = StaticField}), 
                   table);

        emit2(".field", "public static", s ^ " " ^ jvm_type(tipe))
    )

    | gen_field_decl(_, _) = (print("Fatal error: this node is not a field declaration.\n"); raise Halt)
and
      gen_decl(Field(Array(tipe, size), id(name), line, col), table) = 
      let val var = allocate_jvm_var()
      in
          insert_symb(name,
                CodegenInfo({info = Info({name = name, 
                                         tipe = tipe, 
                                         decl = Field(Array(tipe, size), id(name), line, col), 
                                         whatisit = CSXField}),
                             addr = StackVar(var)}), 
                   table);
        conjoin([
          ldc(Int32.toString(size)),
          emit1("newarray", jasmin_type(tipe)),
          emit1("astore", Int.toString(var))
         ])
      end

    | gen_decl(Field(tipe, id(s), line, col), table) = ( 
    let val var = allocate_jvm_var() in (
        insert_symb(s,
                CodegenInfo({info = Info({name = s, 
                                         tipe = tipe, 
                                         decl = Field(tipe, id(s), line, col), 
                                         whatisit = CSXField}),
                             addr = StackVar(var)}), 
                   table);
         conjoin([emit("iconst_0"),
                  istore(var)])
         )end

    )
    | gen_decl(InitializedField(UnknownConst, id(s), value, line, col), table) =
    (print("gen_decl(InitializedField(UnknownConst...))"); raise Bug)
    | gen_decl(InitializedField(tipe, id(s), value, line, col), table) = 
    (
    let val var = allocate_jvm_var() in (
        insert_symb(s,
                CodegenInfo({info = Info({name = s, 
                                         tipe = tipe, 
                                         decl = Field(tipe, id(s), line, col), 
                                         whatisit = CSXField}),
                             addr = StackVar(var)}), 
                   table);
         conjoin([gen_expr_t(value, table),
                  istore(var)])
       ) end
    )
    | gen_decl(FormalArg(ArrayParam(tipe), id(name), line, col), table) = (
        insert_symb(name,
                    CodegenInfo({info = Info({name = name, tipe = tipe, decl = FormalArg(ArrayParam(tipe), id(name), line, col), whatisit = CSXFormalArg}),
                                 addr = StackVar(allocate_jvm_var())}),
                    table);
        ""
        )
    | gen_decl(FormalArg(tipe, id(name), line, col), table) =  (
        insert_symb(name,
                    CodegenInfo({info = Info({name = name, tipe = tipe, decl = FormalArg(tipe, id(name), line, col), whatisit = CSXFormalArg}),
                                 addr = StackVar(allocate_jvm_var())}),
                    table);
        ""
        )
    | gen_decl(MainMethod(field_decls, stmts, line, col), table) = ( 
         jvm_register := 0;
         insert_symb("main",
                 Info({name = "main", 
                                         tipe = Void, 
                                         decl = MainMethod(field_decls, stmts, line, col),
                                         whatisit = CSXMethod}),
                   table);
        let val num_locals = 1 + List.length(field_decls)
            val num_pushes = 1 + push_count_decls(field_decls) + push_count_stmts(stmts)
            val augmented_table = open_scope(table) in
          conjoin([emit2(".method", "public static", "main([Ljava/lang/String;)V"),
                   emit1(".limit locals", Int.toString(num_locals)),
                   emit1(".limit stack", Int.toString(num_pushes)),
                   gen_decls(field_decls, augmented_table),
                   gen_stmts(stmts, augmented_table),
                   emit("return"),
                   emit1(".end", "method")])
        end
    )

    | gen_decl(Method(tipe, id(name), arg_decls, field_decls, stmts, line, col), table) =
    (
         jvm_register := 0;
         let val augmented_table = open_scope(table) in
           insert_symb(name,
                   Info({name = name, 
                         tipe = tipe, 
                         decl = Method(tipe, id(name), arg_decls, field_decls, stmts, line, col),
                         whatisit = CSXMethod}),
                     table);

           let val num_locals = 1 + List.length(field_decls) + List.length(arg_decls)
               val num_pushes = 1 + push_count_decls(field_decls) + push_count_stmts(stmts) 
               val jasmin_arg_decls = 
                              if List.length(arg_decls) > 1 then
                                conjoin(List.map (fn x => jvm_type(arg_type(x))^"") arg_decls)
                              else
                                conjoin(List.map (fn x => jvm_type(arg_type(x))) arg_decls) 
               val return_type = jvm_type(tipe)
               val method_sig = name^"("^jasmin_arg_decls^")"^return_type  in
            gen_decls(arg_decls, augmented_table);
            conjoin([emit2(".method", "public static", method_sig),
                     emit1(".limit locals", Int.toString(num_locals)),
                     emit1(".limit stack", Int.toString(num_pushes)),
                     gen_decls(field_decls, augmented_table),
                     gen_stmts(stmts, augmented_table),
                     emit("return"),
                     emit1(".end", "method")])
           end
        end
    )
and 
     gen_stmts(stmts, table) = conjoin(List.map (fn stmt => gen_stmt(stmt, table)) stmts)
and 
       gen_stmt(NoneStmt, table) = ("") 
     | gen_stmt(Block(decls, stmts, line, col), table) = (
      let val augmented_table = open_scope(table) in
        conjoin([
           gen_decls(decls, augmented_table),
           gen_stmts(stmts, augmented_table)
        ])
      end
     )

     | gen_stmt(IfThen(cond, then_stmt, line, col), table) =
         gen_stmt(IfThenElse(cond, then_stmt, NoneStmt, 0, 0), table)

     | gen_stmt(IfThenElse(cond, then_stmt, else_stmt, line, col), table) = (
       let val else_label = gen_label()
           val end_label = gen_label() in (
        conjoin([
          gen_expr_t(cond, table),
          emit1("ifeq", else_label),
          gen_stmt(then_stmt, table),
          emit1("goto", end_label),
          emit_label(else_label),
          gen_stmt(else_stmt, table),
          emit_label(end_label)
        ])
       ) 
     end )

     | gen_stmt(While(cond, body, line, col), table) = 
       let val top = gen_label()
           val bottom = gen_label() in (
           conjoin([emit_label(top), 
                    gen_expr_t(cond, table), 
                    emit1("ifeq", bottom), 
                    gen_stmt(body, table), 
                    goto(top), 
                    emit_label(bottom)])
       ) end

     | gen_stmt(LabeledWhile(Label(id(name)), cond, body, line, col), table) = 
       let val begin_label = gen_label()
           val end_label = gen_label()
           val augmented_table = add_label(LabelBeginEnd(id(name), begin_label, end_label), table)
       in
           conjoin([emit_label(begin_label), 
                    gen_expr_t(cond, augmented_table), 
                    emit1("ifeq", end_label), 
                    gen_stmt(body, augmented_table), 
                    goto(begin_label), 
                    emit_label(end_label)])
       end

     | gen_stmt(Asg(Checked(tipe, ArrayDeref(id(s), expr, _, _)), rval, line, col), table) = (
     let val static = is_static(s, table) 
         val index = gen_expr_t(expr, table)
         (* what on EARTH. 3? why dummy = 3?  FIXME *)
         val dummy = 3
         val save = array_store_val(tipe)
         val deref = if static then
                       conjoin([get_static(Array(tipe, dummy), s), index]) 
                     else
                       conjoin([emit1("aload", Int.toString(get_var(s, table))), index]) in
       conjoin([
         deref,
         gen_expr_t(rval, table),
         save 
       ])
     end
     )

     | gen_stmt(Asg(Checked(Array(tipe, n), Name(id(s), _, _)),  (* lval *)
                Checked(_, StrLit(the_string, _, _)), line, col),  (* rval *)
                table) = (
     let val static = is_static(s, table) 
         val save = if static then
                      put_static(Array(tipe, n), s)
                    else
                      array_store(tipe, get_var(s, table))
         val load = if static then
                      get_static(Array(tipe, n), s)
                     else
                       emit1("aload", Int.toString(get_var(s, table))) in
       conjoin([
         load,
         ldc(the_string),
         emit1("invokestatic", "CSXLib/convertString(Ljava/lang/String;)[C"),
         (*array_clone(typeof(rval)), (*runtime size check happens in * array_clone*)*)
         save 
       ])
     end
     )

     | gen_stmt(Asg(Checked(ArrayParam(tipeL), Name(id(sL), _, _)), 
                    Checked(ArrayParam(tipeR), Name(id(sR), _, _)), line, col), table) = (
            let val idx = Int.toString(get_var(sL, table)) in (
                 conjoin([emit1("aload", idx),
                          gen_expr_t(Checked(ArrayParam(tipeR), Name(id(sR), 0, 0)), table),
                          array_clone(ArrayParam(tipeR)),
                          emit1("astore", idx)])
            )
            end
     )

     | gen_stmt(Asg(Checked(ArrayParam(tipeL), Name(id(sL), _, _)), 
                    Checked(Array(tipeR, nR), Name(id(sR), _, _)), line, col), table) = (
            let val idx = Int.toString(get_var(sL, table)) in (
                 conjoin([emit1("aload", idx),
                          gen_expr_t(Checked(Array(tipeR, nR), Name(id(sR), 0, 0)), table),
                          array_clone(Array(tipeR, nR)),
                          emit1("astore", idx)])
            )
            end
     )
     
     | gen_stmt(Asg(Checked(Array(tipeL, nL), Name(id(sL), _, _)), 
                    Checked(ArrayParam(tipeR), Name(id(sR), _, _)), line, col), table) = (
         if is_static(sL, table) then (
                 conjoin([get_static(tipeL, sL),
                          gen_expr_t(Checked(ArrayParam(tipeR), Name(id(sR), 0, 0)), table),
                          array_clone(ArrayParam(tipeR)),
                          put_static(Array(tipeL, nL), sL)])
          )
         else
            let val idx = Int.toString(get_var(sL, table)) in (
                 conjoin([emit1("aload", idx),
                          gen_expr_t(Checked(ArrayParam(tipeR), Name(id(sR), 0, 0)), table),
                          array_clone(ArrayParam(tipeR)),
                          emit1("astore", idx)])
            )
            end
     )

     | gen_stmt(Asg(Checked(Array(tipe, n), Name(id(s), _, _)), rval, line, col), table) = (
     let val static = is_static(s, table) 
         val save = if static then
                      put_static(Array(tipe, n), s)
                    else
                      array_store(tipe, get_var(s, table))
         val load = if static then
                      get_static(Array(tipe, n), s)
                     else
                       emit1("aload", Int.toString(get_var(s, table))) in
       conjoin([
         load,
         gen_expr_t(rval, table),
         array_clone(typeof(rval)), (*runtime size check happens in array_clone*)
         save 
       ])
     end
     )


     | gen_stmt(Asg(lval, rval, line, col), table) = (
       conjoin([
         gen_expr_t(rval, table),
         if is_static(name_of(lval), table) then
           (put_static(typeof(lval), name_of(lval)))
         else
           let val register = get_var(name_of(lval), table) in 
             (istore(register))
           end
       ])
     )

     | gen_stmt(Read(expr, line, col), table) =  
       let val static = is_static(name_of(expr), table) in
       conjoin([
         case typeof(expr) of Int => emit("invokestatic CSXLib/readInt()I")
                            | Char => emit("invokestatic CSXLib/readChar()C")
                            | _ => raise UnreadableType,
         if static then put_static(typeof(expr), name_of(expr)) else istore(get_var(name_of(expr), table))
         ])
       end 
                           
     | gen_stmt(Reads(reads, line, col), table) = emit(List.foldr (op ^) "" (List.map (fn r => gen_stmt(r, table)) reads))

     | gen_stmt(Writes(stmts, line, col), table) = emit(List.foldr (op ^) "" (List.map (fn w => gen_stmt(w, table)) stmts))
     | gen_stmt(Write(expr, line, col), table) = (
            let val expr_type = typeof_expr(expr, table) in
                conjoin([
                gen_expr_t(expr, table),
                emit1("invokestatic", print_code_for(expr_type))])
            end
     )
     | gen_stmt(ProcCall(id(name), args, line, col), table) = (
      let val jasmin_arg_decls = conjoin(List.map (fn x => jvm_type(typeof(x))) args) 
          val rt = (get_return_type_of(name, table)) in
        conjoin([gen_actual_args(args, table),
                 call(name, jasmin_arg_decls, rt)
                 ])
      end
      )
     | gen_stmt(ArglessProcCall(id(name), line, col), table) = call(name, "", Void)
     | gen_stmt(Return(NoneExpr, line, col), table) = emit("return")
     | gen_stmt(Return(expr, line, col), table) = conjoin([gen_expr_t(expr,
     table), emit("ireturn")])
     | gen_stmt(Break(Label(id(name)), line, col), table) = (
        let val (_, end_label) = get_label_begin_end(name, table) in
            goto(end_label)
        end
     )
     | gen_stmt(Continue(Label(id(name)), line, col), table) = (
        let val (begin_label, _) = get_label_begin_end(name, table) in
            goto(begin_label)
        end
     )
     | gen_stmt(Exit(code, line, col), table) =  (
       conjoin([ 
         ldc(if code < 0 then "-"^Int32.toString(abs(code)) else Int32.toString(code)),
         emit1("invokestatic", "java/lang/System/exit(I)V")
       ])
       )
and
    gen_rel_binary_op(LT, table) =  "if_icmplt"
  | gen_rel_binary_op(GT, table) =  "if_icmpgt"
  | gen_rel_binary_op(LEQ, table) = "if_icmple"
  | gen_rel_binary_op(GEQ, table) = "if_icmpge"
and
    gen_equal_binary_op(EQ, table) = "if_icmpeq"
  | gen_equal_binary_op(NOTEQ, table) = "if_icmpne"
and
    gen_arit_binary_op(PLUS, table) = emit("iadd")
  | gen_arit_binary_op(MINUS, table) =  emit("isub")
  | gen_arit_binary_op(TIMES, table) = emit("imul")
  | gen_arit_binary_op(DIV, table) = emit("idiv")
and
   gen_logi_binary_op(AND, table) = "iand"
 | gen_logi_binary_op(OR, table) = "ior"
and 
    print_code_for(Char) = "CSXLib/printChar(C)V"
  | print_code_for(Int) = "CSXLib/printInt(I)V"
  | print_code_for(Bool) = "CSXLib/printBool(Z)V"
  | print_code_for(StringLit(n)) = "CSXLib/printString(Ljava/lang/String;)V"
  | print_code_for(Array(Char, n)) = "CSXLib/printCharArray([C)V"
  | print_code_for(Const(t)) = print_code_for(t)
  | print_code_for(Param(t)) = print_code_for(t)
  | print_code_for(ArrayParam(t)) = print_code_for(Array(t, 0))
  | print_code_for(Array(t, n)) = raise NotImplemented 
  | print_code_for(UnknownConst) = raise Bug 
  | print_code_for(Error) = "" 
  | print_code_for(Void) = raise Bug
  | print_code_for(NoneType) = raise Bug

and 
  locals_count(Method(tipe, name, arg_decls, field_decls, stmts, _, _)) = 
     List.length(arg_decls) + List.length(field_decls)
  | locals_count(MainMethod(field_decls, stmts, _, _)) = List.length(field_decls)
  | locals_count(_) = 0
and
    array_store(Int, var) = emit1("astore", Int.toString(var))
  | array_store(Char, var) = emit1("astore", Int.toString(var))
  | array_store(Bool, var) = emit1("astore", Int.toString(var))
  | array_store(_) = raise Halt
and
    array_store_val(Int) = emit("iastore")
  | array_store_val(Char) = emit("castore")
  | array_store_val(Bool) = emit("bastore")
  | array_store_val(_) = raise Halt
and
    array_clone(Array(Int, n)) = conjoin([
      emit1("invokestatic", "CSXLib/checkIntArrayLength([I[I)[I"),
      emit1("invokestatic", "CSXLib/cloneIntArray([I)[I")
    ])
  | array_clone(Array(Char, n)) = conjoin([ 
      emit1("invokestatic", "CSXLib/checkCharArrayLength([C[C)[C"),
      emit1("invokestatic", "CSXLib/cloneCharArray([C)[C")
      ])
  | array_clone(Array(Bool, n)) = conjoin([
      emit1("invokestatic", "CSXLib/checkBoolArrayLength([Z[Z)[Z"),
      emit1("invokestatic", "CSXLib/cloneBoolArray([Z)[Z")
      ])
  | array_clone(ArrayParam(t)) = array_clone(Array(t, 0))
  | array_clone(x) = (print("array_clone("^typeToString(x)^")"); raise Halt)
and
  array_load(var) = emit1("aload", Int.toString(var))
and
  arg_type(FormalArg(tipe, id(s), _, _)) = tipe
 |arg_type(NoneDecl) = NoneType
 |arg_type(_) = raise Halt

end;
