structure TypeCheck =
struct 

open DataTypes;
open SymbolTable;
open Unparser;
open DeclaredMain;

exception Halt;

val any_errors = ref false;

fun println(s : string) = print(s^"\n");

fun everyp(_, nil) = true
  | everyp(p, first::rest) = if p(first) then
                                everyp(p, rest)
                             else
                                false
fun somep(_, nil) = false
  | somep(p, first::rest) = if p(first) then
                                true 
                             else
                                somep(p, rest)

fun errorp(t) = t = Error


fun report_error(message : string, line : int, col : int) = 
        let val line_str = Int.toString(line)
            val col_str  = Int.toString(col) in
           (
           any_errors := true;
           print("Error - "^line_str^":"^col_str^" "^message^"\n"))
        end

fun report_stmt_error(message : string, line : int, col : int) =
    (report_error(message, line, col); 
     NoneStmt)

fun report_decl_error(message : string, line : int, col : int) =
    (report_error(message, line, col); 
     NoneDecl)

fun report_asg_error(Error, _, line, col) = NoneStmt
  | report_asg_error(_, Error, line, col) = NoneStmt
  | report_asg_error(Array(Error, _), _, line, col) = NoneStmt
  | report_asg_error(_, Array(Error, _), line, col) = NoneStmt
  | report_asg_error(ArrayParam(Error), _, line, col) = NoneStmt
  | report_asg_error(_, ArrayParam(Error), line, col) = NoneStmt
  | report_asg_error(Const(lhs_type), _, line, col) = (
     report_error("a const cannot receive assignment", line, col);
     NoneStmt
  )
  | report_asg_error(lhs_type, rhs_type, line, col) = (
     report_error(typeToString(rhs_type)^" cannot be assigned to "^typeToString(lhs_type), line, col);
     NoneStmt
  )


fun report_arg_match_error(lhs_type, rhs_type, line, col) = (
    if somep(errorp, [lhs_type, rhs_type]) then
       ()
    else
      report_error("the formal arg of type ("^typeToString(lhs_type)^") does not match the argument ("^typeToString(rhs_type)^")", 
                   line, col);
    false
 )
    
  (*report_error(typeToString(rhs_type)^" cannot be assigned to *
  * "^typeToString(rhs_type), line, col)*)

fun warn(str) = println("Warning: "^str);

fun already_declared(name, line, col) =  (
  report_error("symbol "^name^" is already declared", line, col);
  NoneDecl
)

fun report_type_error(message, [], line, col) = (report_error(message, line, col); Error)
  | report_type_error(message, types, line, col) = (
              if somep(errorp, types) then
                Error (* suppress message *)
              else
                (report_error(message, line, col); Error)
    )



fun integralp(Const(t)) = integralp(t) 
  | integralp(Char) = true
  | integralp(Int) = true
  | integralp(Bool) = false
  | integralp(x) = ((*print(typeToString(x)^"is not integral");*)false)

fun arithmeticp(Const(t)) = integralp(t) 
  | arithmeticp(Char) = true
  | arithmeticp(Int) = true
  | arithmeticp(Bool) = false 
  | arithmeticp(x) = ((*print(typeToString(x)^"is not integral");*)false)

fun booleanp(Const(t)) = booleanp(t)
  | booleanp(Bool) = true
  | booleanp(_) = false

fun readable(Char) = true
  | readable(Int) = true
  | readable(Const(t)) = false 
  | readable(Array(t, _)) = t = Char
  | readable(_) = false

fun truep(x) = x = true

fun check_class_name(class_name, file_name) =
    if String.isSuffix (class_name^".csx") file_name then
      true
       else
         (report_error("A csx program must end with the csx extension, and the classname must match the filename", 0, 0);
        false)

fun constantp(InitializedField(UnknownConst, id(name), rhs, line, col)) = true
  | constantp(_) = false

fun printable(t) = 
     case t of Int  => true
             | Bool  => true
             | Char  => true
             | Void  => false
             | Array(Int, _) => false
             | ArrayParam(Int) => false
             | Array(Bool, _) => false
             | ArrayParam(Bool) => false
             | Const(t) => printable(t) 
             | Param(t) => printable(t)
             | Array(t, _) => printable(t) 
             | ArrayParam(t) => printable(t) 
             | StringLit(n) => true 
             | UnknownConst => false
             | NoneType => false
             | Error => false (* Error is used in the typechecker. *)


(* fun castable(a, cast(b)) ~= does a cast to b? 
 * eg: 
 * castable(Char, Cast(Int)) = "can a char be cast to an int?"
 *)

fun castable(Char, Cast(Int)) = true
  | castable(Char, Cast(Char)) = true
  | castable(Char, Cast(Bool)) = true
  | castable(Int, Cast(Int)) = true
  | castable(Int, Cast(Char)) = true
  | castable(Int, Cast(Bool)) = true
  | castable(Bool, Cast(Int)) = true
  | castable(Bool, Cast(Char)) = true
  | castable(Bool, Cast(Bool)) = true
  | castable(_, _) = false

fun widest(Char, Int) = Int
  | widest(Int, Char) = Int
  | widest(Char, Char) = Char
  | widest(Int, Int) = Int
  | widest(Error, _) = Error
  | widest(_, Error) = Error
  | widest(Const(t0), Const(t1)) = widest(t0, t1)
  | widest(Const(t0), t1) = widest(t0, t1)
  | widest(t0, Const(t1)) = widest(t0, t1)
  | widest(_, _) = Error 

 (* FIXME: This Yes/No stuff is for the birds.
  * I intended on using Maybe with some warning messages, but that didn't
  * work out very well.  
  * FIXME. just use boolean values.*)
datatype will_accept = Yes | No 

(* accepts(a, b) is Yes when x, of type a, will receive assignment from y 
 * of type b in the statement:  x = y;
 *)
fun accepts(Const(_), _) = No 
  | accepts(x, Const(y)) = accepts(x, y)
  | accepts(Bool, Bool) = Yes
  | accepts(Bool, Int) = No
  | accepts(Bool, Char) = No
  | accepts(Bool, Array(t, n)) = No
  | accepts(Int, Bool) = No
  | accepts(Int, Int) = Yes
  | accepts(Int, Char) = Yes
  | accepts(Int, Array(t, n)) = No 
  | accepts(Char, Bool) = Yes
  | accepts(Char, Int) = Yes
  | accepts(Char, Char) = Yes
  | accepts(Char, Array(t, n)) = No 
  | accepts(UnknownConst, _) = No 
  | accepts(Array(Char, array_len), StringLit(str_len)) = if
  Int32.fromInt(str_len) = array_len then Yes else No
  | accepts(Array(tl, nl), Array(tr, nr)) = if nl = nr andalso tl = tr then Yes else No
  | accepts(ArrayParam(tl), ArrayParam(tr)) = if tl = tr then Yes else No
  | accepts(ArrayParam(tl), Array(tr, nr)) = 
        if tl = tr then 
          (
            (*warn("assigning a local array to a param fails at runtime when * their sizes differ");*)
            Yes
          )
        else
          No
  | accepts(Array(tl, nl), ArrayParam(tr)) = 
        if tl = tr then 
          (
            (*warn("assigning an param to a local array fails at runtime when * * their sizes differ");*)
            Yes
          )
        else
          No 
  | accepts(Array(line, col), _) = No
  | accepts(ArrayParam(_), _) = No
  | accepts(Error, _) = No
  | accepts(_, Error) = No
  | accepts(_, _) = No
  (*| accepts(a, b) = (print("accepts("^typeToString(a)^", "^typeToString(b)^") go boom"); raise Halt; No)*)

(* FIXME, we only do this with arg decls *)
fun decl_accepts(Field(tipe, _, _, _), rhs_type) = accepts(tipe, rhs_type)
  | decl_accepts(InitializedField(Const(t), _, _, _, _), rhs_type) = No
  | decl_accepts(InitializedField(tipe, _, _, _, _), rhs_type) = accepts(tipe, rhs_type)
  | decl_accepts(Method(tipe, _, _, _, _, _, _), rhs_type) = No 
  | decl_accepts(MainMethod(_, _, _, _), rhs_type) = No 
  | decl_accepts(FormalArg(tipe, _, _, _), rhs_type) = accepts(tipe, rhs_type)
  | decl_accepts(NoneDecl, _) = No


fun check(Program(id(name), members), filename) = (
        let 
            val has_main = DeclaredMain.correctly_declared_main(Program(id(name), members))
            val checked_members = check_members(members, make_symbol_table())
            val dummy1 = check_class_name(name, filename) in
          (
            if has_main then
             NoneDecl
            else
             (report_decl_error("the main method must be declared last, and only once.", 1, 1));

            (!any_errors, Program(id(name), checked_members)))
        end

    )
and 
    check_members(MemberDecls(fields, methods), table) = 
    let val checked_fields =  List.map (fn field_decl => check_decl(field_decl, table)) fields
        val checked_methods = List.map (fn method_decl => check_decl(method_decl, table)) methods
    in 
      MemberDecls(checked_fields, checked_methods)
    end  

and 
     check_unary_op(NOT, table) = (NOT)
   | check_unary_op(Cast(tipe), table) = (Cast(tipe))
and 
     check_exprs(exprs, table) = List.map (fn e => check_expr(e, table)) exprs
and 
      rebuild_expr(NoneExpr, table) = (NoneExpr)

    | rebuild_expr(ArithmeticBinaryOperation(lhs, arit_binary_operator, rhs, line, col), table) =
        ArithmeticBinaryOperation(check_expr(lhs, table), 
                                  arit_binary_operator, 
                                  check_expr(rhs, table), 
                                  line, 
                                  col)

    | rebuild_expr(LogicalBinaryOperation(lhs, logi_binary_operator, rhs, line, col), table) = 
        LogicalBinaryOperation(check_expr(lhs, table), 
                               logi_binary_operator, 
                               check_expr(rhs, table), 
                               line, 
                               col)

    | rebuild_expr(RelationalBinaryOperation(lhs, rel_binary_operator, rhs, line, col), table) = 
        RelationalBinaryOperation(check_expr(lhs, table), 
                                  rel_binary_operator, 
                                  check_expr(rhs, table), 
                                  line, 
                                  col)

    | rebuild_expr(EqualityBinaryOperation(lhs, equal_binary_operator, rhs, line, col), table) = 
        EqualityBinaryOperation(check_expr(lhs, table), 
                                equal_binary_operator, 
                                check_expr(rhs, table), 
                                line, 
                                col)

    | rebuild_expr(UnaryOperation (unary_op, expr, line, col), table) =
        UnaryOperation(unary_op, check_expr(expr, table), line, col)

    | rebuild_expr(CharLit(the_char, line, col), table) = CharLit(the_char, line, col)

    | rebuild_expr(IntLit(the_num, line, col), table) = 
       IntLit(the_num, line, col)

    | rebuild_expr(StrLit(the_str, line, col), table) = StrLit(the_str, line, col)

    | rebuild_expr(True(_, _), table) = True(0, 0)

    | rebuild_expr(False(_, _), table) = False(0, 0)

    | rebuild_expr(ArglessFnCall(id(name), line, col), table) =  (
         ArglessFnCall(id(name), line, col)
    )
    | rebuild_expr(FnCall(id(name), actual_args, line, col), table) = (

        let val formal_args = get_method_args(name, table)
            val return_type = typeof_expr(ArglessFnCall(id(name), line, col), table)
            val checked_args = List.map (fn a => check_expr(a, table)) actual_args in
           if List.length(formal_args) = List.length(actual_args) then (
              (ListPair.map 
                   (fn (f, a) => ( 
                            let val actual_arg_t = typeof_expr(a, table)
                                val formal_arg_t = decl_type(f, table) in
                             case decl_accepts(f, actual_arg_t) of Yes => true
                                                                 | No  =>  (
                              report_arg_match_error(formal_arg_t, actual_arg_t, line, col)) 
                            end))
                   (formal_args, actual_args));
              FnCall(id(name), checked_args, line, col)
           ) else (
             report_type_error("wrong number of arguments in proc call", [return_type], line, col);
             NoneExpr
           )
        end
    )

    | rebuild_expr(Name(id(name), line, col), table) = (
        Name(id(name), line, col)
    )

    | rebuild_expr(ArrayDeref(id(name), index, line, col), table) = (
      if integralp(typeof_expr(index, table)) then
        let val t = typeof_expr(Name(id(name), line, col), table) 
            val this = ArrayDeref(id(name), check_expr(index, table), line, col) in
          case t 
             of Array(t, n) => this
              | ArrayParam(t) => this (* what a terrible name *)
              | _ => (report_type_error("the symbol "^name^" is not an array, and cannot be dereferenced", [t], line, col);
                      NoneExpr)
        end
      else
          (report_type_error("array indices can only be of integral type", [typeof_expr(index, table)], line, col);
           NoneExpr)
    )
    | rebuild_expr(ActualArgument(expr, line, col), table) = 
          ActualArgument(check_expr(expr, table), line, col)
    | rebuild_expr(Checked(t, e), table) = Checked(t, e)

  and
      check_expr(expr, table) = 
      let val tipe = typeof_expr(expr, table) in
        Checked(tipe, rebuild_expr(expr, table))
      end

and 
    (* typeof_expr tells us the type of the thing left on the runtime stack. *)
      typeof_expr(NoneExpr, table) = (Void)
    | typeof_expr(ArithmeticBinaryOperation(lhs, arit_binary_operator, rhs, line, col), table) = (
      (* plus, minus, stuff like that. *)
      let val lhs_type = typeof_expr(lhs, table)
          val rhs_type = typeof_expr(rhs, table) 
          val all_arithmetic = everyp(arithmeticp, [lhs_type, rhs_type])
          val should_suppress_msg = somep(fn x => x = Error, [lhs_type, rhs_type]) in
            if  all_arithmetic then
              (
                widest(lhs_type, rhs_type))
            else
                if should_suppress_msg then
                  Error 
                else
              report_type_error("arithmetic operators only apply to integral expressions", [lhs_type, rhs_type], line, col)
      end
    )
    | typeof_expr(LogicalBinaryOperation(lhs, logi_binary_operator, rhs, line, col), table) = (
      (* and, or, stuff like that *)
      let val lhs_type = typeof_expr(lhs, table)
          val rhs_type = typeof_expr(rhs, table) in 
            if everyp(booleanp, [lhs_type, rhs_type]) then (
              Bool
              )
            else
              report_type_error("operands for logical operators must be have the boolean type", [lhs_type, rhs_type], line, col)
      end
    )
    | typeof_expr(RelationalBinaryOperation(lhs, rel_binary_operator, rhs, line, col), table) = (
      (* >, <  stuff like that *)
      let val lhs_type = typeof_expr(lhs, table)
          val rhs_type = typeof_expr(rhs, table) in 
            if everyp(integralp, [lhs_type, rhs_type]) orelse
               everyp(booleanp, [lhs_type, rhs_type]) then
              Bool
            else
              (
              report_type_error("operands for relational operators must be integral", [lhs_type, rhs_type], line, col))
      end
    )
    | typeof_expr(EqualityBinaryOperation(lhs, equal_binary_operator, rhs, line, col), table) = (
      (* ==, !=  *)
      let val lhs_type = typeof_expr(lhs, table)
          val rhs_type = typeof_expr(rhs, table) in 
            if everyp(integralp, [lhs_type, rhs_type]) orelse
               everyp(booleanp, [lhs_type, rhs_type]) then(
               Bool )
            else (
              report_type_error("operands for equality operators must have comparable types", [lhs_type, rhs_type], line, col)
            )
      end
    )
    | typeof_expr(UnaryOperation (Cast(t), expr, line, col), table) =  (
        let val op_t = typeof_expr(expr, table) in 
            if castable(op_t, Cast(t)) then
                t
            else
              let val want_t_str = typeToString(t)
                  val have_t_str = typeToString(op_t) in
                  report_type_error("expressions of type "^have_t_str^" cannot be cast to "^want_t_str, [op_t], line, col)
              end
        end
    )
    | typeof_expr(UnaryOperation (NOT, expr, line, col), table) =  (
        let val op_t = typeof_expr(expr, table) in 
            if op_t = Bool then
                Bool
            else
              report_type_error("bang only works on boolean expressions", [op_t], line, col)
        end
    )

    | typeof_expr(CharLit(the_char, line, col), table) = Char

    | typeof_expr(IntLit(the_num, line, col), table) = (Int)

    | typeof_expr(StrLit(the_str, line, col), table) = StringLit(String.size(the_str) - 2)(*(Array(Char, String.size(the_str) - 2))*)

    | typeof_expr(True(_), table) = (Bool)

    | typeof_expr(False(_), table) = (Bool)

    | typeof_expr(ArglessFnCall(id(name), line, col), table) = (
       let val info = get_symb(name, table) in
        case info of Info({name = name, tipe = tipe, decl = decl, whatisit = CSXMethod}) => tipe
                  |  _ => report_type_error(name^" is not a method and cannot be called", [], line, col)
       end
       handle NotFound => (report_type_error("the method "^name^"() is not defined.", [], line, col))
            | NotAMethod => report_type_error("the name "^name^" doesn't identify a method.", [], line, col)
    )
    | typeof_expr(FnCall(id(name), actual_args, line, col), table) = (
        get_return_type_of(name, table)
       handle NotFound => (report_type_error("the method "^name^"() is not defined.", [], line, col))
            | NotAMethod => report_type_error("the name "^name^" doesn't identify a method.", [], line, col)
    )

    | typeof_expr(Name(id(name), line, col), table) = (
        let val exists = symb_exists(name, table)
            val is_a_method = is_method(name, table) in
                if exists then
                  (
                        if is_a_method then 
                                report_type_error("a method cannot be referenced", [], line, col)
                        else
                                type_of(name, table)
                 )
                else
                  report_type_error("the symbol "^name^" is not defined", [], line, col)
        end
    )

    | typeof_expr(ArrayDeref(id(name), expr, line, col), table) = (
        let val t = typeof_expr(Name(id(name), line, col), table) in
          case t 
             of Array(t, n) => t
              | ArrayParam(t) => t                                                           
              | _ => (report_type_error("the symbol "^name^" is not an array, and cannot be dereferenced", [t], line, col))
        end
    )
    
    | typeof_expr(ActualArgument(expr, line, col), table) =  typeof_expr(expr, table)

    | typeof_expr(Checked(tipe, expr), table) = (tipe)
and 
      discard_const(Const(t)) = t
    | discard_const(t) = t
and
     check_actual_args(args, table) = List.map (fn e =>(check_expr(e); print(", "))) args
and 
    check_decls(decls, table) = List.map (fn d => check_decl(d, table)) decls
and
    check_decl(Field(Array(tipe, size), id(s), line, col), table) = (
    let val thetype = 
        if size < 1 then (
          report_decl_error("an array size must be greater than 0", line, col);
          Error
        )
        else
          tipe
        val dummy = (insert_symb(s,
                                Info({name = s, 
                                      tipe = Array(thetype, size), 
                                      decl = Field(Array(thetype, size), id(s), line, col), whatisit = CSXField}), 
                                    table); NoneDecl
                                     handle DuplicateSymb => already_declared(s, line, col)) in
        Field(Array(thetype, size), id(s), line, col) 
    end
    )
    | check_decl(Field(tipe, id(s), line, col), table) = (
        (insert_symb(s,
                Info({name = s, 
                      tipe = tipe, 
                      decl = Field(tipe, id(s), line, col), 
                      whatisit = CSXField}), 
                    table);
         Field(tipe, id(s), line, col))
        handle DuplicateSymb => already_declared(s, line, col)
    )
    | check_decl(InitializedField(UnknownConst, id(s), value, line, col), table) = (
        let val rhs_tipe = typeof_expr(value, table) in
          insert_symb(s,
                  Info({name = s, 
                        tipe = Const(rhs_tipe), 
                        decl = InitializedField(Const(rhs_tipe), id(s), value, line, col), whatisit = CSXField}), 
                      table);
            InitializedField(Const(rhs_tipe), id(s), check_expr(value, table), line, col)
        end
        handle DuplicateSymb => already_declared(s, line, col)
    )
    
    | check_decl(InitializedField(tipe, id(s), value, line, col), table) = (
      let val checked_rhs =  check_expr(value, table) 
          val Checked(rhs_t, rhs_expr) = checked_rhs in
               (insert_symb(s,
                        Info({name = s, 
                              tipe = tipe, 
                              decl = InitializedField(tipe, id(s), checked_rhs, line, col), whatisit = CSXField}), 
                            table);
                if accepts(tipe, rhs_t) = Yes then
                  InitializedField(tipe, id(s), checked_rhs, line, col)
                else (
                  report_asg_error(tipe, rhs_t, line, col);
                  NoneDecl  
                ))
                handle DuplicateSymb => already_declared(s, line, col)
       end
    )

    | check_decl(FormalArg(ArrayParam(tipe), id(name), line, col), table) = 
    (
          let val dummy = insert_symb(name, 
                                      Info({name = name, 
                                            tipe = ArrayParam(tipe), 
                                            decl = FormalArg(tipe, id(name), line, col), 
                                            whatisit = CSXFormalArg}), 
                                      table) in
            FormalArg(ArrayParam(tipe), id(name), line, col)
          end
          handle DuplicateSymb => already_declared(name, line, col)
    )
    | check_decl(FormalArg(tipe, id(name), line, col), table) = (
          let val dummy = insert_symb(name, 
                                      Info({name = name, tipe = tipe, decl = FormalArg(tipe, id(name), line, col), whatisit = CSXFormalArg}), 
                                      table) in
            FormalArg(tipe, id(name), line, col)
          end
          handle DuplicateSymb => already_declared(name, line, col)
    )
    | check_decl(MainMethod(field_decls, stmts, line, col), table) = (
      (insert_symb("main", 
                   Info({name = "main", 
                         tipe = Void,
                         decl = MainMethod(field_decls, stmts, line, col),
                         whatisit = CSXMethod}),
                     table);
      let val augmented_table = open_scope(table)
          val augmented_table2 = replace_current_method(augmented_table, CurrentMethod(id("main"))) in
          MainMethod(check_decls(field_decls, augmented_table2), 
                     check_stmts(stmts, augmented_table2), line, col)
      end)
      handle DuplicateSymb => already_declared("main", line, col)
    )
    | check_decl(Method(tipe, id(name), arg_decls, field_decls, stmts, line, col), table) = (
      let val dummy = insert_symb(name, 
                                 Info({name = name, 
                                       tipe = tipe,
                                       decl = Method(tipe, id(name), arg_decls, field_decls, stmts, line, col),
                                       whatisit = CSXMethod}),
                                   table) in 
        NoneDecl 
      end
      handle DuplicateSymb => already_declared(name, line, col);


      let val augmented_table = open_scope(table)
          val augmented_table2 = replace_current_method(augmented_table, CurrentMethod(id(name))) 
          val rv = Method(tipe, id(name), check_decls(arg_decls, augmented_table2), 
                                          check_decls(field_decls, augmented_table2), 
                                          check_stmts(stmts, augmented_table2), line, col) in
          if tipe = Void then
            rv
          else
            if returns_unconditionally(stmts) then
              rv
            else
              (report_error("method "^name^" does not return on all possible paths", line, col);
             NoneDecl)
      end
    )
    | check_decl(NoneDecl, table) = NoneDecl
and 
     check_stmts(stmts, table) = List.map (fn stmt => check_stmt(stmt, table)) stmts
and 
       check_stmt(NoneStmt, table) = NoneStmt
     | check_stmt(Block(decls, stmts, line, col), table) = (
        let val augmented_table = open_scope(table) in
          Block(List.map (fn d => check_decl(d, augmented_table)) decls, 
                List.map (fn d => check_stmt(d, augmented_table)) stmts, 
                line, 
                col)
        end
     )
     | check_stmt(IfThen(cond, then_stmt, line, col), table) = (
        check_stmt(IfThenElse(cond, then_stmt, NoneStmt, line, col), table)
     )
     | check_stmt(IfThenElse(cond, then_stmt, else_stmt, line, col), table) = (
         
     let val checked_cond = check_expr(cond, table)
         val Checked(cond_t, cond_expr) = checked_cond 
         val checked_then = check_stmt(then_stmt, table) 
         val checked_else = check_stmt(else_stmt, table)
         val rv = IfThenElse(checked_cond, checked_then, checked_else, line, col) in
        case discard_const(cond_t)
                  of Bool => rv
                   | Error => NoneStmt
                   | _ => (report_stmt_error("the condition must have boolean type", line, col); 
                           NoneStmt)
     end
     )
     | check_stmt(While(cond, stmt, line, col), table) = (
     let val checked_cond = check_expr(cond, table)
         val Checked(cond_t, cond_expr) = checked_cond 
         val checked_stmt = check_stmt(stmt, table) 
         val rv = While(checked_cond, checked_stmt, line, col) in
        case discard_const(cond_t)
                  of Bool => rv
                   | Error => NoneStmt
                   | _ => (report_stmt_error("the condition must have boolean type", line, col); 
                           NoneStmt)
     end
     )
     | check_stmt(LabeledWhile(Label(id(name)), cond, stmt, line, col), table) = (
       if symb_exists(name, table) then
         report_stmt_error("the name "^name^" is already defined.", line, col)
       else
         open_scope_and_then(table, 
               fn newtable => (
               insert_symb(name, CSXLabel, newtable);
               let val checked_cond = check_expr(cond, newtable)
                   val checked_stmt = check_stmt(stmt, newtable)
                   val Checked(cond_t, cond_expr) = checked_cond in
                               case discard_const(cond_t)
                                of Error => NoneStmt
                                 | Bool => LabeledWhile(Label(id(name)), 
                                                        checked_cond,
                                                        checked_stmt, 
                                                        line, 
                                                        col)
                                 | _ => report_stmt_error("the condition must have boolean type", line, col)
                   end
             )
         )
     )
     | check_stmt(LabeledWhile(_, cond, stmt, line, col), table) = raise Halt
     | check_stmt(Asg(Name(id(name), l, c), rval, line, col), table) = ((
       if label_exists(name, table) then 
           report_stmt_error("a label cannot receive assignment", line, col)
       else if is_const(get_symb(name, table)) then 
           report_stmt_error("a const value cannot receive assignment", line, col)

       else if is_method(name, table) then
           report_stmt_error("a method cannot receive assignment", line, col)
       else (
       let val rhs_type = typeof_expr(rval, table) 
           val symbol = get_symb(name, table) 
           val lhs_type = typeof_expr(Name(id(name), l, c), table) in
             (
             if accepts(lhs_type, rhs_type) = Yes  then
                Asg(check_expr(Name(id(name), l, c), table), check_expr(rval, table), line, col)
             else
               report_asg_error(lhs_type, rhs_type, line, col)
               )
       end
       ))
       handle NotFound => report_stmt_error("the name "^name^" is not defined", line, col)
     )
     | check_stmt(Asg(ArrayDeref(id(name), index, l, c), rval, line, col), table) = (
          let val rhs = check_expr(rval, table) 
              val rhs_type = typeof_expr(rval, table)
              val lhs = check_expr(ArrayDeref(id(name), index, l, c), table)
              val lhs_type = typeof_expr(ArrayDeref(id(name), index, l, c), table) in
                case accepts(lhs_type, rhs_type) of Yes => Asg(lhs, rhs, line, col)
                                                  | No  => report_asg_error(lhs_type, rhs_type, line, col)
          end
     )

     | check_stmt(Read(expr, line, col), table) = (
            Read(check_expr(expr, table), line, col)
     )
     | check_stmt(Reads(reads, line, col), table) = (
        Reads(List.map (fn r => check_read(r, table)) reads, line, col)
     )
     | check_stmt(Writes(writes, line, col), table) = (
        Writes(List.map (fn s => check_stmt(s, table)) writes, line, col)
     )
     | check_stmt(Write(expr, line, col), table) = (
       let val tipe = typeof_expr(expr, table) in
         if printable(tipe) then (
          Write(check_expr(expr, table), line, col)
          )
         else
           if tipe = Error then (
             NoneStmt
             )
           else (
            report_stmt_error("values of type "^typeToString(tipe)^" cannot be printed", line, col)
            )
       end
     )
     | check_stmt(ArglessProcCall(id(name), line, col), table) = (
        (* FIXME!  FnCalls look like proc calls.  distinguish by the return type. *)

        if is_method(name, table) then (
          if get_return_type_of(name, table) = Void then 
                ArglessProcCall(id(name), line, col) 
          else 
                report_stmt_error(name^" is a function, and cannot be used as a statement", line, col)
        )
        else (
          report_stmt_error(name^" is not a method", line, col)
        )
     )

     | check_stmt(ProcCall(id(name), actual_args, line, col), table) = (
        if is_method(name, table) then
              if get_return_type_of(name, table) = Void then (
                      let val formal_args = get_method_args(name, table) in
                       if List.length(formal_args) = List.length(actual_args) then (
                          (ListPair.map 
                               (fn (f, a) => ( 
                                  let val actual_arg_t = typeof_expr(a, table)
                                      val formal_arg_t = decl_type(f, table) in
                                         case decl_accepts(f, actual_arg_t) of Yes => true
                                                                             | No  =>  (
                                          report_arg_match_error(formal_arg_t , actual_arg_t, line, col))
                                  end))
                          (formal_args, actual_args));
                          ProcCall(id(name), check_exprs(actual_args, table), line, col)
                       ) else (
                         report_stmt_error("wrong number of arguments in proc call", line, col);
                         NoneStmt
                       )
                      end
              )
              else (
                 report_stmt_error(name^" is a function, and cannot be used as a statement", line, col)
              )
        else
         report_stmt_error(name^" is not a method", line, col)
     )

     | check_stmt(Return(NoneExpr, line, col), table) = (
        let val rt = get_current_return_type(table) in (
          if rt = Void then (
            Return(NoneExpr, line, col)
            )
          else (
            report_stmt_error("must return a value from a non-void method", line, col)
            )
          )
        end
     )

     | check_stmt(Return(expr, line, col), table) = (
        let val return_type = get_current_return_type(table)
            val checked_expr= check_expr(expr, table) 
            val expr_type = typeof_expr(expr, table) in
              if return_type = Void then 
                report_stmt_error("a value cannot be returned from a void method", line, col)
              else if return_type = expr_type then
                Return(checked_expr, line, col)
              else
                report_stmt_error("a returned expression's type must match the return type", line, col)
        end
     )

     | check_stmt(Break(Label(id(name)), line, col), table) = (
         let val lab = check_label(Label(id(name)), line, col, table) in
            Break(lab, line, col)
         end
     )

     | check_stmt(Break(_, line, col), table) = raise Halt

     | check_stmt(Continue(Label(id(name)), line, col), table) = (
         let val label = check_label(Label(id(name)), line, col, table)
         in
           Continue(label, line, col)
         end
     )

     | check_stmt(Continue(_, line, col), table) = raise Halt

     | check_stmt(Exit(code, line, col), table) = (
        Exit(code, line, col)
     )
and
     check_label(Label(id(name)), line, col, table) = (
        if label_exists(name, table) then
          Label(id(name))
        else (
            report_error("the name "^name^" does not identify a label", line, col);
            NoneLabel
        )
     )
     | check_label(LabelBeginEnd(_, _, _), line, col, table) = raise Halt
     | check_label(NoneLabel, line, col, table) = raise Halt
and
    check_rela_binary_op(LT, table) = LT
  | check_rela_binary_op(GT, table) = GT
  | check_rela_binary_op(LEQ, table) = LEQ
  | check_rela_binary_op(GEQ, table) = GEQ
and
    check_equal_binary_op(EQ, table) = EQ
  | check_equal_binary_op(NOTEQ, table) = NOTEQ
and
    check_arit_binary_op(PLUS, table) = PLUS
  | check_arit_binary_op(MINUS, table) = MINUS
  | check_arit_binary_op(TIMES, table) = TIMES
  | check_arit_binary_op(DIV, table) = DIV
and
   check_logi_binary_op(AND, table) = AND
 | check_logi_binary_op(OR, table) = OR
and
     returns_unconditionally(stmts) =  somep(truep, List.map (fn stmt => returns(stmt)) stmts)
and 
     returns(NoneStmt) = false
  |  returns(Block(_, stmts, _, _)) = returns_unconditionally(stmts)
  |  returns(IfThen(_, then_part, _, _)) = false
  |  returns(IfThenElse(_, then_part, else_part, _, _)) = returns(then_part) andalso returns(else_part)
  |  returns(While(_, body, _, _)) = returns(body)
  |  returns(LabeledWhile(_, _, body, _, _)) = returns(body)
  |  returns(Asg(_, _, _, _)) = false
  |  returns(Read(_, _, _)) = false
  |  returns(Reads(_, _, _)) = false
  |  returns(Write(_, _, _)) = false
  |  returns(Writes(_, _, _)) = false
  |  returns(ArglessProcCall(_, _, _)) = false
  |  returns(ProcCall(_, _, _, _)) = false 
  |  returns(Return(_, _, _)) = true 
  |  returns(Break(_, _, _)) = false
  |  returns(Continue(_, _, _)) = false
  |  returns(Exit(_, _, _)) = true
and
  check_read(Read(expr, line, col), table) = (
     case typeof_expr(expr, table) of NoneType => report_stmt_error("labels cannot be read", line, col)
                                    | t => if readable(t) then
                                             check_stmt(Read(expr, line, col), table)
                                           else
                                             report_stmt_error("the type " ^typeToString(t)^" cannot be read", line, col)
  )
                     
end;
